@@ -24,6 +24,11 @@
 #include <vector>
 #include <sys/types.h>
 #include <dirent.h>
+#include <getopt.h>      // getopt_long()
+
+#ifdef HAVE_GRP_H
+#  include <grp.h>       // initgroups()
+#endif
 
 /* bring in inet_ntop if it is not present */
 #define ETH_ALEN 6
@@ -31,6 +36,16 @@
 #include "inet_ntop.c"
 #endif
 
+#ifdef HAVE_CAP_NG_H
+#include <cap-ng.h>
+#endif
+
+/* droproot is from tcpdump.
+ * See https://github.com/the-tcpdump-group/tcpdump/blob/master/tcpdump.c#L611
+ */
+const char *program_name = 0;
+const char *tcpflow_droproot_username = 0;
+const char *tcpflow_chroot_dir = 0;
 
 scanner_info::scanner_config be_config; // system configuration
 
@@ -53,7 +68,7 @@ const char *progname = 0;		// name of the program
 int debug = DEFAULT_DEBUG_LEVEL;	// global variable, not clear why
 
 /* semaphore prevents multiple copies from outputing on top of each other */
-#ifdef HAVE_PTHREAD
+#ifdef HAVE_PTHREAD_H
 #include <semaphore.h>
 sem_t *semlock = 0;
 #endif
@@ -61,13 +76,14 @@ sem_t *semlock = 0;
 #define DEFAULT_REPORT_FILENAME "report.xml"
 
 /****************************************************************
- *** SCANNER PLUG-IN SYSTEM 
+ *** SCANNER PLUG-IN SYSTEM
  ****************************************************************/
 
 scanner_t *scanners_builtin[] = {
     scan_md5,
     scan_http,
     scan_netviz,
+    scan_python,
     scan_tcpdemux,
 #ifdef USE_WIFI
     scan_wifiviz,
@@ -76,6 +92,22 @@ scanner_t *scanners_builtin[] = {
 
 bool opt_no_promisc = false;		// true if we should not use promiscious mode
 
+/* Long options!
+ *
+ * We need more long options; developers looking at this file should
+ * feel free to submit more!
+ */
+
+static const struct option longopts[] = {
+    { "chroot", required_argument, NULL, 'z' },
+    { "help", no_argument, NULL, 'h' },
+    { "relinquish-privileges", required_argument, NULL, 'U' },
+    { "verbose", no_argument, NULL, 'v' },
+    { "version", no_argument, NULL, 'V' },
+    { NULL, 0, NULL, 0 }
+};
+
+
 /****************************************************************
  *** USAGE
  ****************************************************************/
@@ -83,11 +115,11 @@ bool opt_no_promisc = false;		// true if we should not use promiscious mode
 static void usage(int level)
 {
     std::cout << PACKAGE_NAME << " version " << PACKAGE_VERSION << "\n\n";
-    std::cout << "usage: " << progname << " [-aBcCDhJpsvVZ] [-b max_bytes] [-d debug_level] \n";
-    std::cout << "     [-[eE] scanner] [-f max_fds] [-F[ctTXMkmg]] [-i iface] [-L semlock]\n";
-    std::cout << "     [-m min_bytes] [-o outdir] [-r file] [-R file] \n";
-    std::cout << "     [-S name=value] [-T template] [-w file] [-x scanner] [-X xmlfile]\n";
-    std::cout << "      [expression]\n\n";
+    std::cout << "usage: " << progname << " [-aBcCDhIpsvVZ] [-b max_bytes] [-d debug_level] \n";
+    std::cout << "     [-[eE] scanner] [-f max_fds] [-F[ctTXMkmg]] [-h|--help] [-i iface]\n";
+    std::cout << "     [-l files...] [-L semlock] [-m min_bytes] [-o outdir] [-r file] [-R file]\n";
+    std::cout << "     [-S name=value] [-T template] [-U|--relinquish-privileges user] [-v|--verbose]\n";
+    std::cout << "     [-w file] [-x scanner] [-X xmlfile] [-z|--chroot dir] [expression]\n\n";
     std::cout << "   -a: do ALL post-processing.\n";
     std::cout << "   -b max_bytes: max number of bytes per flow to save\n";
     std::cout << "   -d debug_level: debug level; default is " << DEFAULT_DEBUG_LEVEL << "\n";
@@ -95,17 +127,18 @@ static void usage(int level)
     std::cout << "   -h: print this help message (-hh for more help)\n";
     std::cout << "   -H: print detailed information about each scanner\n";
     std::cout << "   -i: network interface on which to listen\n";
-    std::cout << "   -I: generate temporal packet-> byte index files for each flow (.findex)\n";
+    std::cout << "   -I: write for each flow another file *.findx to provide byte-indexed timestamps\n";
     std::cout << "   -g: output each flow in alternating colors (note change!)\n";
     std::cout << "   -l: treat non-flag arguments as input files rather than a pcap expression\n";
     std::cout << "   -L  semlock - specifies that writes are locked using a named semaphore\n";
     std::cout << "   -p: don't use promiscuous mode\n";
     std::cout << "   -q: quiet mode - do not print warnings\n";
-    std::cout << "   -r file: read packets from tcpdump pcap file (may be repeated)\n";
-    std::cout << "   -R file: read packets from tcpdump pcap file TO FINISH CONNECTIONS\n";
-    std::cout << "   -v: verbose operation equivalent to -d 10\n";
-    std::cout << "   -V: print version number and exit\n";
-    std::cout << "   -w file: write packets not processed to file\n";
+
+    std::cout << "   -r file      : read packets from tcpdump pcap file (may be repeated)\n";
+    std::cout << "   -R file      : read packets from tcpdump pcap file TO FINISH CONNECTIONS\n";
+    std::cout << "   -v           : verbose operation equivalent to -d 10\n";
+    std::cout << "   -V           : print version number and exit\n";
+    std::cout << "   -w  file     : write packets not processed to file\n";
     std::cout << "   -o  outdir   : specify output directory (default '.')\n";
     std::cout << "   -X  filename : DFXML output to filename\n";
     std::cout << "   -m  bytes    : specifies skip that starts a new stream (default "
@@ -113,18 +146,29 @@ static void usage(int level)
     std::cout << "   -F{p} : filename prefix/suffix (-hh for options)\n";
     std::cout << "   -T{t} : filename template (-hh for options; default "
               << flow::filename_template << ")\n";
-    std::cout << "   -Z: do not decompress gzip-compressed HTTP transactions\n";
+    std::cout << "   -Z       do not decompress gzip-compressed HTTP transactions\n";
+
+    std::cout << "\nSecurity:\n";
+    std::cout << "   -U user  relinquish privleges and become user (if running as root)\n";
+    std::cout << "   -z dir   chroot to dir (requires that -U be used).\n";
 
     std::cout << "\nControl of Scanners:\n";
     std::cout << "   -E scanner   - turn off all scanners except scanner\n";
     std::cout << "   -S name=value  Set a configuration parameter (-hh for info)\n";
-    be13::plugin::info_scanners(false,true,scanners_builtin,'e','x');
-
-    std::cout << "Console output options:\n";
+    if(level > 1) {
+        std::cout << "\n" "Activated options -S name=value:";
+        for(int i=0;defaults[i].name;i++){
+            std::cout <<"\n   -S "<< defaults[i].name << "=" << defaults[i].dvalue <<'\t'<< defaults[i].help;
+        }
+        std::cout << '\n';
+        be13::plugin::info_scanners(false,true,scanners_builtin,'e','x');
+    }
+    std::cout << "\n"
+                 "Console output options:\n";
     std::cout << "   -B: binary output, even with -c or -C (normally -c or -C turn it off)\n";
     std::cout << "   -c: console print only (don't create files)\n";
     std::cout << "   -C: console print only, but without the display of source/dest header\n";
-    std::cout << "   -0: don't print newlines after packets when printing to console";
+    std::cout << "   -0: don't print newlines after packets when printing to console\n";
     std::cout << "   -s: strip non-printable characters (change to '.')\n";
     std::cout << "   -D: output in hex (useful to combine with -c or -C)\n";
     std::cout << "\n";
@@ -136,24 +180,18 @@ static void usage(int level)
     if(level<2) return;
     std::cout << "Filename Prefixes:\n";
     std::cout << "   -Fc : append the connection counter to ALL filenames\n";
-    std::cout << "   -Ft : prepend the time_t timestamp to ALL filenames\n";
-    std::cout << "   -FT : prepend the ISO8601 timestamp to ALL filenames\n";
+    std::cout << "   -Ft : prepend the time_t UTC timestamp to ALL filenames\n";
+    std::cout << "   -FT : prepend the ISO8601 UTC timestamp to ALL filenames\n";
     std::cout << "   -FX : Do not output any files (other than report files)\n";
     std::cout << "   -FM : Calculate the MD5 for every flow (stores in DFXML)\n";
     std::cout << "   -Fk : Bin output in 1K directories\n";
     std::cout << "   -Fm : Bin output in 1M directories (2 levels)\n";
     std::cout << "   -Fg : Bin output in 1G directories (3 levels)\n";
     flow::usage();
-    std::cout << "-S name=value options:\n";
-    for(int i=0;defaults[i].name;i++){
-        std::stringstream ss;
-        ss << defaults[i].name << "=" << defaults[i].dvalue;
-        printf("   %-20s %s\n",ss.str().c_str(),defaults[i].help);
-    }
-    std::cout << "\n";
-    std::cout << "DEBUG Levels (specify with -dNN):\n";
-    std::cout << "get_max_fds() = " << tcpdemux::getInstance()->get_max_fds() << "\n";
-    std::cout << "NUM_RESERVED_FDS = " << NUM_RESERVED_FDS << "\n";
+    std::cout << "\n" "Current limitations:"
+                 "\n" "  get_max_fds() = " << tcpdemux::getInstance()->get_max_fds();
+    std::cout << "\n" "  NUM_RESERVED_FDS = " << NUM_RESERVED_FDS;
+    std::cout << '\n';
 }
 
 /**
@@ -196,14 +234,22 @@ void replace(std::string &str,const std::string &from,const std::string &to)
 /* These must be global variables so they are available in the signal handler */
 feature_recorder_set *the_fs = 0;
 dfxml_writer *xreport = 0;
+pcap_t *pd = 0;
 void terminate(int sig)
 {
-    DEBUG(1) ("terminating");
-    be13::plugin::phase_shutdown(*the_fs);	// give plugins a chance to do a clean shutdown
-    exit(0); /* libpcap uses onexit to clean up */
+    if (sig == SIGHUP || sig == SIGINT || sig == SIGTERM) {
+        DEBUG(1) ("terminating orderly");
+        pcap_breakloop(pd);
+        return;
+    } else {
+        DEBUG(1) ("terminating");
+        be13::plugin::phase_shutdown(*the_fs);	// give plugins a chance to do a clean shutdown
+        exit(0); /* libpcap uses onexit to clean up */
+    }
 }
 
 #ifdef HAVE_FORK
+#include <sys/wait.h>
 // transparent decompression for process_infile
 class inflater {
     const std::string suffix;
@@ -217,7 +263,7 @@ public:
         return ends_with(file_path,suffix);
     }
     // invoke the inflater in a shell, and return the file descriptor to read the inflated file from
-    int invoke(const std::string &file_path) const
+    int invoke(const std::string &file_path, int* ppid) const
     {
         std::string invocation = ssprintf(invoc_format.c_str(), file_path.c_str());
         int pipe_fds[2];
@@ -246,6 +292,7 @@ public:
             }
             exit(0);
         }
+        *ppid = child_pid;
         close(pipe_fds[1]);
         return pipe_fds[0];
     }
@@ -273,20 +320,109 @@ static inflaters_t *build_inflaters()
 #define HAVE_INFLATER
 #endif
 
+// https://github.com/the-tcpdump-group/tcpdump/blob/master/tcpdump.c#L611
+#ifndef _WIN32
+/* Drop root privileges and chroot if necessary */
+static void
+droproot(tcpdemux &demux,const char *username, const char *chroot_dir)
+{
+    struct passwd *pw = NULL;
+
+    if (chroot_dir && !username) {
+        fprintf(stderr, "%s: Chroot without dropping root is insecure\n",
+                program_name);
+        exit(1);
+    }
+
+    pw = getpwnam(username);
+    if (pw) {
+        /* Begin tcpflow add */
+        if(demux.xreport){
+            const char *outfilename = demux.xreport->get_outfilename().c_str();
+            if(chown(outfilename,pw->pw_uid,pw->pw_gid)){
+                fprintf(stderr, "%s: Coudln't change owner of '%.64s' to %s (uid %d): %s\n",
+                        program_name, outfilename, username, pw->pw_uid, strerror(errno));
+                exit(1);
+            }
+        }
+        /* end tcpflow add */
+        if (chroot_dir) {
+            if (chroot(chroot_dir) != 0 || chdir ("/") != 0) {
+                fprintf(stderr, "%s: Couldn't chroot/chdir to '%.64s': %s\n",
+                        program_name, chroot_dir, pcap_strerror(errno));
+                exit(1);
+            }
+        }
+#ifdef HAVE_LIBCAP_NG
+        {
+            int ret = capng_change_id(pw->pw_uid, pw->pw_gid, CAPNG_NO_FLAG);
+            if (ret < 0) {
+                fprintf(stderr, "error : ret %d\n", ret);
+            } else {
+                fprintf(stderr, "dropped privs to %s\n", username);
+            }
+        }
+#else
+        if (initgroups(pw->pw_name, pw->pw_gid) != 0 ||
+            setgid(pw->pw_gid) != 0 || setuid(pw->pw_uid) != 0) {
+            fprintf(stderr, "%s: Couldn't change to '%.32s' uid=%lu gid=%lu: %s\n",
+                    program_name, username,
+                    (unsigned long)pw->pw_uid,
+                    (unsigned long)pw->pw_gid,
+                    pcap_strerror(errno));
+            exit(1);
+        }
+        else {
+            fprintf(stderr, "dropped privs to %s\n", username);
+        }
+#endif /* HAVE_LIBCAP_NG */
+    }
+    else {
+        fprintf(stderr, "%s: Couldn't find user '%.32s'\n",
+                program_name, username);
+        exit(1);
+    }
+#ifdef HAVE_LIBCAP_NG
+    /* We don't need CAP_SETUID, CAP_SETGID and CAP_SYS_CHROOT any more. */
+    capng_updatev(
+                  CAPNG_DROP,
+                  (capng_type_t)(CAPNG_EFFECTIVE | CAPNG_PERMITTED),
+                  CAP_SETUID,
+                  CAP_SETGID,
+                  CAP_SYS_CHROOT,
+                  -1);
+    capng_apply(CAPNG_SELECT_BOTH);
+#endif /* HAVE_LIBCAP_NG */
+
+}
+#endif /* _WIN32 */
+
+/**
+ * Perform the droproot operation for tcpflow. This needs to be called immediately after pcap_open()
+ */
+void tcpflow_droproot(tcpdemux &demux)
+{
+    if (tcpflow_droproot_username){
+        droproot(demux,tcpflow_droproot_username,tcpflow_chroot_dir);
+    }
+}
+
 /*
  * process an input file or device
  * May be repeated.
  * If start is false, do not initiate new connections
+ * Return 0 on success or -1 on error
  */
 #ifdef HAVE_INFLATER
 static inflaters_t *inflaters = 0;
 #endif
-static void process_infile(const std::string &expression,const char *device,const std::string &infile)
+static int process_infile(tcpdemux &demux,const std::string &expression,const char *device,const std::string &infile)
 {
     char error[PCAP_ERRBUF_SIZE];
-    pcap_t *pd=0;
     int dlt=0;
     pcap_handler handler;
+    int waitfor = -1;
+    int pipefd = -1;
 
 #ifdef HAVE_INFLATER
     if(inflaters==0) inflaters = build_inflaters();
@@ -298,12 +434,12 @@ static void process_infile(const std::string &expression,const char *device,cons
 #ifdef HAVE_INFLATER
         for(inflaters_t::const_iterator it = inflaters->begin(); it != inflaters->end(); it++) {
             if((*it)->appropriate(infile)) {
-                int fd = (*it)->invoke(infile);
-                file_path = ssprintf("/dev/fd/%d", fd);
-                if(fd < 0) {
-                    std::cerr << "decompression of '" << infile << "' failed" << std::endl;
+                pipefd = (*it)->invoke(infile, &waitfor);
+                if(pipefd < 0) {
+                    std::cerr << "decompression of '" << infile << "' failed: " << strerror (errno) << std::endl;
                     exit(1);
                 }
+                file_path = ssprintf("/dev/fd/%d", pipefd);
                 if(access(file_path.c_str(), R_OK)) {
                     std::cerr << "decompression of '" << infile << "' is not available on this system" << std::endl;
                     exit(1);
@@ -315,6 +451,7 @@ static void process_infile(const std::string &expression,const char *device,cons
 	if ((pd = pcap_open_offline(file_path.c_str(), error)) == NULL){	/* open the capture file */
 	    die("%s", error);
 	}
+        tcpflow_droproot(demux);        // drop root if requested
 	dlt = pcap_datalink(pd);	/* get the handler for this kind of packets */
 	handler = find_handler(dlt, infile.c_str());
     } else {
@@ -329,12 +466,7 @@ static void process_infile(const std::string &expression,const char *device,cons
 	if ((pd = pcap_open_live(device, SNAPLEN, !opt_no_promisc, 1000, error)) == NULL){
 	    die("%s", error);
 	}
-#if defined(HAVE_SETUID) && defined(HAVE_GETUID)
-	/* drop root privileges - we don't need them any more */
-	if(setuid(getuid())){
-	    perror("setuid");
-	}
-#endif
+        tcpflow_droproot(demux);                     // drop root if requested
 	/* get the handler for this kind of packets */
 	dlt = pcap_datalink(pd);
 	handler = find_handler(dlt, device);
@@ -365,10 +497,23 @@ static void process_infile(const std::string &expression,const char *device,cons
 
     /* start listening or reading from the input file */
     if (infile == "") DEBUG(1) ("listening on %s", device);
-    if (pcap_loop(pd, -1, handler, (u_char *)tcpdemux::getInstance()) < 0){
-	
-	die("%s: %s", infile.c_str(),pcap_geterr(pd));
+    int pcap_retval = pcap_loop(pd, -1, handler, (u_char *)tcpdemux::getInstance());
+    
+    if (pcap_retval < 0 && pcap_retval != -2){
+	DEBUG(1) ("%s: %s", infile.c_str(),pcap_geterr(pd));
+	return -1;
     }
+    pcap_close (pd);
+#ifdef HAVE_FORK
+    if (waitfor != -1) {
+        wait (0);
+    }
+    if (pipefd != -1) {
+        close (pipefd);
+    }
+#endif
+
+    return 0;
 }
 
 
@@ -396,6 +541,7 @@ static feature_recorder_set::hash_def be_hash(be_hash_name,be_hash_func);
 
 int main(int argc, char *argv[])
 {
+    program_name = argv[0];
     int opt_help = 0;
     int opt_Help = 0;
     feature_recorder::set_main_threadid();
@@ -411,7 +557,6 @@ int main(int argc, char *argv[])
     bool force_binary_output = false;
     const char *device = 0;             // default device
     const char *lockname = 0;
-    int need_usage = 0;
     std::string reportfilename;
     std::vector<std::string> Rfiles;	// files for finishing
     std::vector<std::string> rfiles;	// files to read
@@ -435,7 +580,7 @@ int main(int argc, char *argv[])
 
     bool trailing_input_list = false;
     int arg;
-    while ((arg = getopt(argc, argv, "aA:Bb:cCd:DE:e:E:F:f:gHhIi:lL:m:o:pqR:r:S:sT:Vvw:x:X:Z:0")) != EOF) {
+    while ((arg = getopt_long(argc, argv, "aA:Bb:cCd:DE:e:E:F:f:gHhIi:lL:m:o:pqR:r:S:sT:U:Vvw:x:X:z:Z0", longopts, NULL)) != EOF) {
 	switch (arg) {
 	case 'a':
 	    demux.opt.post_processing = true;
@@ -443,9 +588,8 @@ int main(int argc, char *argv[])
             be13::plugin::scanners_enable_all();
 	    break;
 
-	case 'A': 
+	case 'A':
 	    fprintf(stderr,"-AH has been deprecated. Just use -a\n");
-	    need_usage=true;
 	    break;
 
 	case 'b':
@@ -484,7 +628,7 @@ int main(int argc, char *argv[])
 	    break;
         case 'e':
             be13::plugin::scanners_enable(optarg);
-            demux.opt.post_processing = true; // enable post processing if anything is turned on 
+            demux.opt.post_processing = true; // enable post processing if anything is turned on
             break;
 	case 'F':
 	    for(const char *cc=optarg;*cc;cc++){
@@ -499,7 +643,6 @@ int main(int argc, char *argv[])
 		case 'M': demux.opt.opt_md5 = true;break;
 		default:
 		    fprintf(stderr,"-F invalid format specification '%c'\n",*cc);
-		    need_usage = true;
 		}
 	    }
 	    break;
@@ -552,11 +695,13 @@ int main(int argc, char *argv[])
                 flow::filename_template += std::string("%C%c"); // append %C%c if not present
             }
             break;
+        case 'U': tcpflow_droproot_username = optarg; break;
 	case 'V': std::cout << PACKAGE_NAME << " " << PACKAGE_VERSION << "\n"; exit (1);
 	case 'v': debug = 10; break;
         case 'w': opt_unk_packets = optarg;break;
 	case 'x': be13::plugin::scanners_disable(optarg);break;
 	case 'X': reportfilename = optarg;break;
+        case 'z': tcpflow_chroot_dir = optarg; break;
 	case 'Z': demux.opt.gzip_decompress = 0; break;
 	case 'H': opt_Help += 1; break;
 	case 'h': opt_help += 1; break;
@@ -567,7 +712,10 @@ int main(int argc, char *argv[])
 	}
     }
 
-    
+    if(tcpflow_chroot_dir && !tcpflow_droproot_username){
+        err(1,"-z option requires -U option");
+    }
+
     argc -= optind;
     argv += optind;
 
@@ -588,7 +736,7 @@ int main(int argc, char *argv[])
         usage(opt_help);
         exit(0);
     }
-            
+
 
     if(demux.opt.post_processing && !demux.opt.store_output){
         std::cerr << "ERROR: post_processing currently requires storing output.\n";
@@ -622,12 +770,12 @@ int main(int argc, char *argv[])
 
     /* was a semaphore provided for the lock? */
     if(lockname){
-#if defined(HAVE_SEMAPHORE_H) && defined(HAVE_PTHREAD)
+#if defined(HAVE_SEMAPHORE_H) && defined(HAVE_PTHREAD_H)
 	semlock = sem_open(lockname,O_CREAT,0777,1); // get the semaphore
 #else
 	fprintf(stderr,"%s: attempt to create lock pthreads not present\n",argv[0]);
 	exit(1);
-#endif	
+#endif
     }
 
     if(force_binary_output) demux.opt.output_strip_nonprint = false;
@@ -653,8 +801,15 @@ int main(int argc, char *argv[])
         }
     }
 
-    /* report file specified? */
+    /* report file specified? If so, open it.
+     * Note: If we are going to chroot, we need apply the chroot prefix also,
+     * but we need to open the file *now*.
+     */
     if(reportfilename.size()>0 && opt_enable_report){
+        if (tcpflow_chroot_dir){
+            reportfilename = std::string(tcpflow_chroot_dir) + std::string("/") + reportfilename;
+        }
+        std::cerr << "reportfilename: " << reportfilename << "\n";
 	xreport = new dfxml_writer(reportfilename,false);
 	dfxml_create(*xreport,command_line);
 	demux.xreport = xreport;
@@ -699,31 +854,35 @@ int main(int argc, char *argv[])
 
 
     /* Process r files and R files */
+    int exit_val = 0;
     if(xreport){
         xreport->push("configuration");
     }
     if(rfiles.size()==0 && Rfiles.size()==0){
 	/* live capture */
-#if defined(HAVE_SETUID) && defined(HAVE_GETUID)
-        /* Since we don't need network access, drop root privileges */
-	if(setuid(getuid())){
-	    perror("setuid");
-	}
-#endif
 	demux.start_new_connections = true;
-        process_infile(expression,device,"");
+        int err = process_infile(demux,expression,device,"");
+        if (err < 0) {
+            exit_val = 1;
+        }
         input_fname = device;
     }
     else {
 	/* first pick up the new connections with -r */
 	demux.start_new_connections = true;
 	for(std::vector<std::string>::const_iterator it=rfiles.begin();it!=rfiles.end();it++){
-	    process_infile(expression,device,*it);
+	    int err = process_infile(demux,expression,device,*it);
+	    if (err < 0) {
+	        exit_val = 1;
+	    }
 	}
 	/* now pick up the outstanding connection with -R, but don't start new connections */
 	demux.start_new_connections = false;
 	for(std::vector<std::string>::const_iterator it=Rfiles.begin();it!=Rfiles.end();it++){
-	    process_infile(expression,device,*it);
+	    int err = process_infile(demux,expression,device,*it);
+	    if (err < 0) {
+	        exit_val = 1;
+	    }
 	}
     }
 
@@ -737,7 +896,7 @@ int main(int argc, char *argv[])
     int open_fds = (int)demux.open_flows.size();
     int flow_map_size = (int)demux.flow_map.size();
 
-    demux.close_all_fd();
+    demux.remove_all_flows();	// empty the map to capture the state
     std::stringstream ss;
     be13::plugin::phase_shutdown(fs,xreport ? &ss : 0);
 
@@ -746,13 +905,11 @@ int main(int argc, char *argv[])
      */
     const std::string total_flow_processed("Total flows processed: %" PRId64);
     const std::string total_packets_processed("Total packets processed: %" PRId64);
-    
+
     DEBUG(2)(total_flow_processed.c_str(),demux.flow_counter);
     DEBUG(2)(total_packets_processed.c_str(),demux.packet_counter);
 
     if(xreport){
-
-	demux.remove_all_flows();	// empty the map to capture the state
         xreport->pop();                 // fileobjects
         xreport->xmlout("summary",ss.str(),"",false);
         xreport->xmlout("open_fds_at_end",open_fds);
@@ -763,7 +920,7 @@ int main(int argc, char *argv[])
 	xreport->add_rusage();
 	xreport->pop();                 // bulk_extractor
 	xreport->close();
-	delete xreport;                 
+	delete xreport;
     }
 
     if(demux.flow_counter > tcpdemux::WARN_TOO_MANY_FILES){
@@ -797,5 +954,5 @@ int main(int argc, char *argv[])
         }
     }
 
-    exit(0);                            // return(0) causes crash on Windows
+    exit(exit_val);                     // return(0) causes crash on Windows
 }
