@@ -2,6 +2,11 @@
  * This file is part of tcpflow by Simson Garfinkel,
  * originally by Jeremy Elson <jelson@circlemud.org>
  *
+ * Modified by Greg Drew to add support for creating a packet time / data index
+ * which allows mapping bytes in the flow back to their relative arrival time.
+ * This is very useful in reassembling inherently bidirectional conversations
+ * such as chat or telnet sessions.  --GDD
+ *
  * This source code is under the GNU Public License (GPL).  See
  * LICENSE for details.
  *
@@ -13,12 +18,12 @@
 
 #include <iostream>
 #include <sstream>
+#include <vector>
+#include <string>
 
 #pragma GCC diagnostic ignored "-Weffc++"
 #pragma GCC diagnostic ignored "-Wshadow"
 
-static int ct=0;
-
 
 /* Create a new tcp object.
  * 
@@ -31,6 +36,7 @@ tcpip::tcpip(tcpdemux &demux_,const flow &flow_,be13::tcp_seq isn_):
     demux(demux_),myflow(flow_),dir(unknown),isn(isn_),nsn(0),
     syn_count(0),fin_count(0),fin_size(0),pos(0),
     flow_pathname(),fd(-1),file_created(false),
+    flow_index_pathname(),idx_file(),
     seen(new recon_set()),
     last_byte(),
     last_packet_number(),out_of_order_count(0),violations(0)
@@ -69,17 +75,18 @@ void tcpip::dump_xml(class dfxml_writer *xreport,const std::string &xmladd)
     std::stringstream attrs;
     attrs << "startime='" << dfxml_writer::to8601(myflow.tstart) << "' ";
     attrs << "endtime='"  << dfxml_writer::to8601(myflow.tlast)  << "' ";
-    attrs << "src_ipn='"  << myflow.src << "' ";
-    attrs << "dst_ipn='"  << myflow.dst << "' ";
     if(myflow.has_mac_daddr()) attrs << "mac_daddr='" << macaddr(myflow.mac_daddr) << "' ";
     if(myflow.has_mac_saddr()) attrs << "mac_saddr='" << macaddr(myflow.mac_saddr) << "' ";
-    attrs << "packets='"  << myflow.packet_count << "' ";
+    attrs << "family='"   << (int)myflow.family << "' ";
+    attrs << "src_ipn='"  << ipaddr_prn(myflow.src, myflow.family) << "' ";
+    attrs << "dst_ipn='"  << ipaddr_prn(myflow.dst, myflow.family) << "' ";
     attrs << "srcport='"  << myflow.sport << "' ";
     attrs << "dstport='"  << myflow.dport << "' ";
-    attrs << "family='"   << (int)myflow.family << "' ";
+    attrs << "packets='"  << myflow.packet_count << "' ";
     if(out_of_order_count) attrs << "out_of_order_count='" << out_of_order_count << "' ";
     if(violations)         attrs << "violations='" << violations << "' ";
-	
+    attrs << "len='"      << myflow.len << "' ";
+    if(myflow.len != myflow.caplen) attrs << "caplen='"   << myflow.caplen << "' ";
     xreport->xmlout(tcpflow_str,"",attrs.str(),false);
     if(xmladd.size()>0) xreport->xmlout("",xmladd,"",false);
     xreport->pop();
@@ -120,12 +127,6 @@ tcpip::~tcpip()
  */
 void tcpip::close_file()
 {
-    ct++;
-    //std::cerr << "close_file0 " << ct << " " << *this << "\n";
-    if(ct==122){
-        //std::cerr << "ct==122\n";
-    }
-
     if (fd>=0){
 	struct timeval times[2];
 	times[0] = myflow.tstart;
@@ -150,8 +151,12 @@ void tcpip::close_file()
 #endif
 	close(fd);
 	fd = -1;
+	demux.open_flows.erase(this);           // we are no longer open
+    }
+    // Also close the flow_index file, if flow indexing is in use --GDD
+    if(demux.opt.output_packet_index && idx_file.is_open()){
+    	idx_file.close();
     }
-    demux.open_flows.erase(this);           // we are no longer open
     //std::cerr << "close_file1 " << *this << "\n";
 }
 
@@ -163,13 +168,14 @@ void tcpip::close_file()
 
 int tcpip::open_file()
 {
-    ct++;
+	int create_idx_needed = false;
     if(fd<0){
         //std::cerr << "open_file0 " << ct << " " << *this << "\n";
         /* If we don't have a filename, create the flow */
         if(flow_pathname.size()==0) {
-            flow_pathname = myflow.new_filename(&fd,O_RDWR|O_BINARY|O_CREAT,0666);
+            flow_pathname = myflow.new_filename(&fd,O_RDWR|O_BINARY|O_CREAT|O_EXCL,0666);
             file_created = true;		// remember we made it
+            create_idx_needed = true;	// We created a new stream, so we need to create a new flow file. --GDD
             DEBUG(5) ("%s: created new file",flow_pathname.c_str());
         } else {
             /* open an existing flow */
@@ -187,10 +193,34 @@ int tcpip::open_file()
             return -1;
         }
         /* Remember that we have this open */
-        demux.open_flows.insert(this);
+        demux.open_flows.push_back(this);
         if(demux.open_flows.size() > demux.max_open_flows) demux.max_open_flows = demux.open_flows.size();
         //std::cerr << "open_file1 " << *this << "\n";
     }
+    if(demux.opt.output_packet_index){
+    	//Open the file for the flow index.  We don't do this if the flow file could not be
+    	//	opened.  The file must be opened for append, in case this is a reopen.  The filename
+    	//	standard is the flow name followed by ".findx", which google currently says does not
+    	//	conflict with anything major.
+    	flow_index_pathname = flow_pathname + ".findx";
+    	DEBUG(10)("opening index file: %s",flow_index_pathname.c_str());
+    	if(create_idx_needed){
+    		//New flow file, even if there was an old one laying around --GDD
+    		idx_file.open(flow_index_pathname.c_str(),std::ios::trunc|std::ios::in|std::ios::out);
+    	}else{
+    		//Use existing flow file --GDD
+    		idx_file.open(flow_index_pathname.c_str(),std::ios::ate|std::ios::in|std::ios::out);
+    	}
+    	if(idx_file.bad()){
+    		perror(flow_index_pathname.c_str());
+    		// Be nice and be sure the flow has been closed in the demultiplexer.
+    		// demux.close_tcpip_fd(this);  Need to fix this.  Also, when called, it will
+    		// have to differentiate the fact that the open fd cound only needs to be
+    		// decremented by one and not by 2.--GDD
+    		return -1;
+    	}
+
+    }
     return 0;
 }
 
@@ -207,10 +237,12 @@ void tcpip::print_packet(const u_char *data, uint32_t length)
     /* green, blue, read */
     const char *color[3] = { "\033[0;32m", "\033[0;34m", "\033[0;31m" };
 
-    if(demux.opt.max_bytes_per_flow>0){
-	if(last_byte > demux.opt.max_bytes_per_flow) return; /* too much has been printed */
-	if(length > demux.opt.max_bytes_per_flow - last_byte){
-	    length = demux.opt.max_bytes_per_flow - last_byte; /* can only output this much */
+    if(demux.opt.max_bytes_per_flow>=0){
+        uint64_t max_bytes_per_flow = (uint64_t)demux.opt.max_bytes_per_flow;
+
+	if(last_byte > max_bytes_per_flow) return; /* too much has been printed */
+	if(length > max_bytes_per_flow - last_byte){
+	    length = max_bytes_per_flow - last_byte; /* can only output this much */
 	    if(length==0) return;
 	}
     }
@@ -240,8 +272,10 @@ void tcpip::print_packet(const u_char *data, uint32_t length)
             
             /* Print the offset */
             char b[64];
-            int count = snprintf(b,sizeof(b),"%04x: ",(int)i);
-            fwrite(b,1,count,stdout);
+            size_t count = snprintf(b,sizeof(b),"%04x: ",(int)i);
+            if(fwrite(b,1,count,stdout)!=count){
+	      perror("fwrite");
+	    }
             spaces += count;
             
             /* Print the hext bytes */
@@ -272,20 +306,28 @@ void tcpip::print_packet(const u_char *data, uint32_t length)
     else if(demux.opt.output_strip_nonprint){
 	for(const u_char *cc = data;cc<data+length;cc++){
 	    if(isprint(*cc) || (*cc=='\n') || (*cc=='\r')){
-		written += fputc(*cc,stdout);
+                int ret = fputc(*cc,stdout);
+                if(ret==EOF){
+                    std::cerr << "EOF on write to stdout\n";
+                    exit(1);
+                
+                }
 	    }
+	    else fputc('.',stdout);
+            written += 1; // treat even unprintable characters as "written". It
+                          // really means "processed"
 	}
     }
     else {
 	written = fwrite(data,1,length,stdout);
+        if(length != written) std::cerr << "\nwrite error to stdout (" << length << "!=" << written << ") \n";
     }
-    if(length != written) std::cerr << "\nwrite error to stdout\n";
 
     last_byte += length;
 
     if (demux.opt.use_color) printf("\033[0m");
 
-    putchar('\n');
+    if (! demux.opt.console_output_nonewline) putchar('\n');
     fflush(stdout);
 
 #ifdef HAVE_PTHREAD
@@ -356,7 +398,7 @@ void update_seen(recon_set *seen,uint64_t pos,uint32_t length)
  *
  * called from tcpdemux::process_tcp_packet()
  */
-void tcpip::store_packet(const u_char *data, uint32_t length, int32_t delta)
+void tcpip::store_packet(const u_char *data, uint32_t length, int32_t delta,struct timeval ts)
 {
     if(length==0) return;               // no need to do anything
 
@@ -369,7 +411,7 @@ void tcpip::store_packet(const u_char *data, uint32_t length, int32_t delta)
 	 * or else we never saw a SYN and we got the ISN wrong.
 	 */
 	if(syn_count>0){
-	    DEBUG(2)("packet received with offset %"PRId64"; ignoring",offset);
+	    DEBUG(2)("packet received with offset %" PRId64 "; ignoring",offset);
 	    violations++;
 	    return;
 	}
@@ -381,13 +423,15 @@ void tcpip::store_packet(const u_char *data, uint32_t length, int32_t delta)
      * but remember to seek out to the actual position after the truncated write...
      */
     uint32_t wlength = length;		// length to write
-    if (demux.opt.max_bytes_per_flow){
-	if(offset >= demux.opt.max_bytes_per_flow){
+    if (demux.opt.max_bytes_per_flow >= 0){
+        uint64_t max_bytes_per_flow = (uint64_t)demux.opt.max_bytes_per_flow;
+
+	if(offset >= max_bytes_per_flow){
 	    wlength = 0;
 	} 
-	if(offset < demux.opt.max_bytes_per_flow &&  offset+length > demux.opt.max_bytes_per_flow){
+	if(offset < max_bytes_per_flow &&  offset+length > max_bytes_per_flow){
 	    DEBUG(2) ("packet truncated by max_bytes_per_flow on %s", flow_pathname.c_str());
-	    wlength = demux.opt.max_bytes_per_flow - offset;
+	    wlength = max_bytes_per_flow - offset;
 	}
     }
 
@@ -396,7 +440,7 @@ void tcpip::store_packet(const u_char *data, uint32_t length, int32_t delta)
      * save the return value because open_tcpfile() puts the file pointer
      * into the structure for us.
      */
-    if (fd < 0 && wlength>0) {
+    if (fd < 0) {
 	if (open_file()) {
 	    DEBUG(1)("unable to open TCP file %s  fd=%d  wlength=%d",
                      flow_pathname.c_str(),fd,(int)wlength);
@@ -413,7 +457,7 @@ void tcpip::store_packet(const u_char *data, uint32_t length, int32_t delta)
 	pos = 0;
 	nsn = isn+1;
 	out_of_order_count++;
-	DEBUG(25)("%s: insert(0,%d); lseek(%d,0,SEEK_SET) out_of_order_count=%"PRId64,
+	DEBUG(25)("%s: insert(0,%d); lseek(%d,0,SEEK_SET) out_of_order_count=%" PRId64,
 		  flow_pathname.c_str(), insert_bytes,
 		  fd,out_of_order_count);
 
@@ -426,16 +470,22 @@ void tcpip::store_packet(const u_char *data, uint32_t length, int32_t delta)
 
     /* if we're not at the correct point in the file, seek there */
     if (offset != pos) {
+        /* Check for a keepalive */
+        if(delta == -1 && length == 1) {
+            DEBUG(25)("%s: RFC1122 keepalive detected and ignored",flow_pathname.c_str());
+            return;
+        }
+
 	if(fd>=0) lseek(fd,(off_t)delta,SEEK_CUR);
 	if(delta<0) out_of_order_count++; // only increment for backwards seeks
-	DEBUG(25)("%s: lseek(%d,%d,SEEK_CUR) offset=%"PRId64" pos=%"PRId64" out_of_order_count=%"PRId64,
+	DEBUG(25)("%s: lseek(%d,%d,SEEK_CUR) offset=%" PRId64 " pos=%" PRId64 " out_of_order_count=%" PRId64,
 		  flow_pathname.c_str(), fd,(int)delta,offset,pos,out_of_order_count);
 	pos += delta;			// where we are now
 	nsn += delta;			// what we expect the nsn to be now
     }
     
     /* write the data into the file */
-    DEBUG(25) ("%s: %s write %ld bytes @%"PRId64,
+    DEBUG(25) ("%s: %s write %ld bytes @%" PRId64,
                flow_pathname.c_str(),
                fd>=0 ? "will" : "won't",
                (long) wlength, offset);
@@ -445,9 +495,20 @@ void tcpip::store_packet(const u_char *data, uint32_t length, int32_t delta)
 	    DEBUG(1) ("write to %s failed: ", flow_pathname.c_str());
 	    if (debug >= 1) perror("");
 	}
+	// Write to the index file if needed.  Note, index file is sorted before close, so no need to jump around --GDD
+		if (demux.opt.output_packet_index && idx_file.is_open()) {
+			idx_file << offset << "|" << ts.tv_sec << "." << std::setw(6) << std::setfill('0') << ts.tv_usec << "|"
+					<< wlength << "\n";
+			if (idx_file.bad()){
+				DEBUG(1)("write to index file %s failed: ",flow_index_pathname.c_str());
+				if(debug >= 1){
+					perror("");
+				}
+			}
+		}
 	if(wlength != length){
 	    off_t p = lseek(fd,length-wlength,SEEK_CUR); // seek out the space we didn't write
-            DEBUG(100)("   lseek(%"PRId64",SEEK_CUR)=%"PRId64,(int64_t)(length-wlength),(int64_t)p);
+            DEBUG(100)("   lseek(%" PRId64 ",SEEK_CUR)=%" PRId64,(int64_t)(length-wlength),(int64_t)p);
 	}
     }
 
@@ -462,7 +523,7 @@ void tcpip::store_packet(const u_char *data, uint32_t length, int32_t delta)
 
     if(debug>=100){
         uint64_t rpos = lseek(fd,(off_t)0,SEEK_CUR);
-        DEBUG(100)("    pos=%"PRId64"  lseek(fd,0,SEEK_CUR)=%"PRId64,pos,rpos);
+        DEBUG(100)("    pos=%" PRId64 "  lseek(fd,0,SEEK_CUR)=%" PRId64,pos,rpos);
         assert(pos==rpos);
     }
 
@@ -472,6 +533,65 @@ void tcpip::store_packet(const u_char *data, uint32_t length, int32_t delta)
 #endif
 }
 
+/*
+ * Compare two index strings and return the result.  Called by
+ * the vector::sort in sort_index.
+ * --GDD
+ */
+bool tcpip::compare(std::string a, std::string b){
+	std::stringstream ss_a(a),ss_b(b);
+	long a_l,b_l;
+
+	ss_a >> a_l;
+	ss_b >> b_l;
+	return a_l < b_l;
+}
+
+/*
+ * Sort an index file (presumably from this object) if file indexing is
+ * turned on and the file exists.  Index files may be out of order due
+ * to the arrival of out of order packets.  It is cheaper to reorder them
+ * one time at the end of processing than it is to continually keep them
+ * in order.
+ * --GDD
+ */
+void tcpip::sort_index(std::fstream *ix_file) {
+
+	std::vector<std::string> idx;
+	std::string line;
+
+	if (demux.opt.output_packet_index) {
+		if (!(idx_file.good() && idx_file.is_open())) {
+			DEBUG(5)("Skipping index file sort.  Unusual behavior.\n");
+			return; //Nothing to do
+		}
+		//Make sure we are at the beginning.
+		ix_file->clear();
+		ix_file->seekg(0);
+		do {
+			*ix_file >> line;
+			if (!ix_file->eof()) {
+				idx.push_back(line);
+			}
+		} while (ix_file->good());
+		std::sort(idx.begin(), idx.end(), &tcpip::compare);
+		ix_file->clear();
+		ix_file->seekg(0);
+		for (std::vector<std::string>::iterator s = idx.begin(); s != idx.end();
+				s++) {
+			*ix_file << *s << "\n";
+		}
+	}
+}
+
+/*
+ * Convenience function to cause the local index file to be sorted.
+ * --GDD
+ */
+void tcpip::sort_index(){
+	tcpip::sort_index(&(this->idx_file));
+}
+
 #pragma GCC diagnostic ignored "-Weffc++"
 #pragma GCC diagnostic ignored "-Wshadow"
 
